
"""
A converter of python values to TOML Token instances.
"""
import codecs
import datetime
import six
import strict_rfc3339
import timestamp
from contoml import tokens
import re
from contoml.elements.metadata import NewlineElement
from contoml.errors import TOMLError


class NotPrimitiveError(TOMLError):
    pass


_operator_tokens_by_type = {
    tokens.TYPE_OP_SQUARE_LEFT_BRACKET: tokens.Token(tokens.TYPE_OP_SQUARE_LEFT_BRACKET, '['),
    tokens.TYPE_OP_SQUARE_RIGHT_BRACKET: tokens.Token(tokens.TYPE_OP_SQUARE_RIGHT_BRACKET, ']'),
    tokens.TYPE_OP_DOUBLE_SQUARE_LEFT_BRACKET: tokens.Token(tokens.TYPE_OP_DOUBLE_SQUARE_LEFT_BRACKET, '[['),
    tokens.TYPE_OP_DOUBLE_SQUARE_RIGHT_BRACKET: tokens.Token(tokens.TYPE_OP_DOUBLE_SQUARE_RIGHT_BRACKET, ']]'),
    tokens.TYPE_OP_COMMA: tokens.Token(tokens.TYPE_OP_COMMA, ','),
    tokens.TYPE_NEWLINE: tokens.Token(tokens.TYPE_NEWLINE, '\n'),
}


def operator_token(token_type):
    return _operator_tokens_by_type[token_type]


_newline = NewlineElement([operator_token(tokens.TYPE_NEWLINE)])


def newline_element():
    return _newline


def create_primitive_token(value):
    """
    Creates and returns a single token for the given primitive atomic value.

    Raises NotPrimitiveError when the given value is not a primitive atomic value
    """
    if isinstance(value, bool):
        return tokens.Token(tokens.TYPE_BOOLEAN, 'true' if value else 'false')
    elif isinstance(value, int):
        return tokens.Token(tokens.TYPE_INTEGER, '{}'.format(value))
    elif isinstance(value, float):
        return tokens.Token(tokens.TYPE_FLOAT, '{}'.format(value))
    elif isinstance(value, (datetime.datetime, datetime.date, datetime.time)):
        ts = timestamp(value) // 1000
        return tokens.Token(tokens.TYPE_DATE, strict_rfc3339.timestamp_to_rfc3339_utcoffset(ts))
    elif isinstance(value, six.string_types):
        return _create_string_token(value)

    raise NotPrimitiveError("{} of type {}".format(value, type(value)))


_bare_string_regex = re.compile('^[a-zA-Z0-9]*$')

def _create_string_token(text):
    if text == '':
        return tokens.Token(tokens.TYPE_STRING, '""'.format(_escape_string(text)))
    elif _bare_string_regex.match(text):
        return tokens.Token(tokens.TYPE_BARE_STRING, text)
    else:
        return tokens.Token(tokens.TYPE_STRING, '"{}"'.format(_escape_string(text)))

def _escape_string(text):
    if six.PY2:
        return text.encode('unicode-escape').encode('string-escape').replace('"', '\\"').replace("\\'", "'")
    else:
        return codecs.encode(text, 'unicode-escape').decode().replace('"', '\\"')





from contoml.elements import common, factory, traversal
from contoml.elements.common import Element, ContainerElement
from contoml.elements.factory import create_element


class ArrayElement(ContainerElement, traversal.TraversalMixin):
    """
    A sequence-like container element containing other atomic elements or other containers.

    Implements list-like interface.

    Assumes input sub_elements are correct for an array element.
    """

    def __init__(self, sub_elements):
        common.ContainerElement.__init__(self, sub_elements)

    def __len__(self):
        return len(tuple(self._enumerate_non_metadata_sub_elements()))

    def __getitem__(self, i):
        """
        Returns the ith entry, which can be a primitive value, a seq-lie, or a dict-like object.
        """
        return self._find_value(i)[1].value

    def __setitem__(self, i, value):
        value_i, _ = self._find_value(i)
        new_element = value if isinstance(value, Element) else factory.create_element(value)
        self._sub_elements = self.sub_elements[:value_i] + [new_element] + self.sub_elements[value_i+1:]

    @property
    def value(self):
        return self     # self is a sequence-like value

    @property
    def primitive_value(self):
        """
        Returns a primitive Python value without any formatting or markup metadata.
        """
        return list(
            self[i].primitive_value if hasattr(self[i], 'primitive_value')
            else self[i]
            for i in range(len(self)))

    def __str__(self):
        return "Array{}".format(self.primitive_value)

    def append(self, v):
        new_entry = [create_element(v)]

        if self:    # If not empty, we need a comma and whitespace prefix!
            new_entry = [
                factory.create_operator_element(','),
                factory.create_whitespace_element(),
            ] + new_entry

        insertion_index = self._find_closing_square_bracket()
        self._sub_elements = self._sub_elements[:insertion_index] + new_entry + \
                             self._sub_elements[insertion_index:]

    def _find_value(self, i):
        """
        Returns (value_index, value) of ith value in this sequence.

        Raises IndexError if not found.
        """
        return tuple(self._enumerate_non_metadata_sub_elements())[i]

    def __delitem__(self, i):
        value_i, value = self._find_value(i)

        begin, end = value_i, value_i+1

        # Rules:
        #   1. begin should be index to the preceding comma to the value
        #   2. end should be index to the following comma, or the closing bracket
        #   3. If no preceding comma found but following comma found then end should be the index of the following value

        preceding_comma = self._find_preceding_comma(value_i)
        found_preceding_comma = preceding_comma >= 0
        if found_preceding_comma:
            begin = preceding_comma

        following_comma = self._find_following_comma(value_i)
        if following_comma >= 0:
            if not found_preceding_comma:
                end = self._find_following_non_metadata(following_comma)
            else:
                end = following_comma
        else:
            end = self._find_closing_square_bracket()

        self._sub_elements = self.sub_elements[:begin] + self._sub_elements[end:]





def is_sequence_like(x):
    """
    Returns True if x exposes a sequence-like interface.
    """
    required_attrs = (
        '__len__',
        '__getitem__'
    )
    return all(hasattr(x, attr) for attr in required_attrs)


def is_dict_like(x):
    """
    Returns True if x exposes a dict-like interface.
    """
    required_attrs = (
        '__len__',
        '__getitem__',
        'keys',
        'values',
    )
    return all(hasattr(x, attr) for attr in required_attrs)


def join_with(iterable, separator):
    """
    Joins elements from iterable with separator and returns the produced sequence as a list.

    separator must be addable to a list.
    """
    inputs = list(iterable)
    b = []
    for i, element in enumerate(inputs):
        if isinstance(element, (list, tuple, set)):
            b += tuple(element)
        else:
            b += [element]
        if i < len(inputs)-1:
            b += separator
    return b



import datetime
import six
from contoml import tokens
from contoml.tokens import py2toml
from contoml.elements.atomic import AtomicElement
from contoml.elements.metadata import PunctuationElement, WhitespaceElement, NewlineElement
from contoml.elements.tableheader import TableHeaderElement
from contoml.util import join_with


def create_element(value):
    """
    Creates and returns the appropriate elements.Element instance from the given Python primitive, sequence-like,
    or dict-like value.
    """
    from contoml.elements.array import ArrayElement
    from contoml.elements.inlinetable import InlineTableElement

    if isinstance(value, (int, float, bool, datetime.datetime, datetime.date) + six.string_types):
        primitive_token = py2toml.create_primitive_token(value)
        return AtomicElement((primitive_token,))

    elif isinstance(value, (list, tuple)):
        preamble = [create_operator_element('[')]
        postable = [create_operator_element(']')]
        stuffing_elements = [create_element(v) for v in value]
        spaced_stuffing = join_with(stuffing_elements,
                                    separator=[create_operator_element(','), create_whitespace_element()])

        return ArrayElement(preamble + spaced_stuffing + postable)

    elif isinstance(value, dict):
        preamble = [create_operator_element('{')]
        postable = [create_operator_element('}')]

        stuffing_elements = (
            (
                create_element(k),
                create_whitespace_element(),
                create_operator_element('='),
                create_whitespace_element(),
                create_element(v)
            ) for (k, v) in value.items())

        spaced_elements = join_with(stuffing_elements,
                                    separator=[create_operator_element(','), create_whitespace_element()])

        return InlineTableElement(preamble + spaced_elements + postable)

    else:
        raise RuntimeError('Value type unaccounted for: {} of type {}'.format(value, type(value)))


def create_operator_element(operator):
    """
    Creates a PunctuationElement instance containing an operator token of the specified type. The operator
    should be a TOML source str.
    """
    operator_type_map = {
        ',': tokens.TYPE_OP_COMMA,
        '=': tokens.TYPE_OP_ASSIGNMENT,
        '[': tokens.TYPE_OP_SQUARE_LEFT_BRACKET,
        ']': tokens.TYPE_OP_SQUARE_RIGHT_BRACKET,
        '[[': tokens.TYPE_OP_DOUBLE_SQUARE_LEFT_BRACKET,
        ']]': tokens.TYPE_OP_DOUBLE_SQUARE_RIGHT_BRACKET,
        '{': tokens.TYPE_OP_CURLY_LEFT_BRACKET,
        '}': tokens.TYPE_OP_CURLY_RIGHT_BRACKET,
    }

    ts = (tokens.Token(operator_type_map[operator], operator),)
    return PunctuationElement(ts)


def create_newline_element():
    """
    Creates and returns a single NewlineElement.
    """
    ts = (tokens.Token(tokens.TYPE_NEWLINE, '\n'),)
    return NewlineElement(ts)


def create_whitespace_element(length=1):
    """
    Creates and returns a WhitespaceElement containing spaces.
    """
    ts = (tokens.Token(tokens.TYPE_WHITESPACE, ' ' * length),)
    return WhitespaceElement(ts)


def create_table_header_element(name):
    return TableHeaderElement((
        py2toml.operator_token(tokens.TYPE_OP_SQUARE_LEFT_BRACKET),
        py2toml.create_primitive_token(name),
        py2toml.operator_token(tokens.TYPE_OP_SQUARE_RIGHT_BRACKET),
        py2toml.operator_token(tokens.TYPE_NEWLINE),
    ))

def create_array_of_tables_header_element(name):
    return TableHeaderElement((
        py2toml.operator_token(tokens.TYPE_OP_DOUBLE_SQUARE_LEFT_BRACKET),
        py2toml.create_primitive_token(name),
        py2toml.operator_token(tokens.TYPE_OP_DOUBLE_SQUARE_RIGHT_BRACKET),
        py2toml.operator_token(tokens.TYPE_NEWLINE),
    ))


def create_table(dict_value):
    """
    Creates a TableElement out of a dict instance.
    """
    from contoml.elements.table import TableElement

    if not isinstance(dict_value, dict):
        raise ValueError('input must be a dict instance.')

    table_element = TableElement([])
    for k, v in dict_value.items():
        table_element[k] = create_element(v)

    return table_element



import datetime

import strict_rfc3339

from contoml import tokens
from contoml.tokens import py2toml


def test_string():
    assert py2toml.create_primitive_token('fawzy') == tokens.Token(tokens.TYPE_BARE_STRING, 'fawzy')
    assert py2toml.create_primitive_token('I am a "cr\'azy" sentence.') == \
           tokens.Token(tokens.TYPE_STRING, '"I am a \\"cr\'azy\\" sentence."')

def test_int():
    assert py2toml.create_primitive_token(42) == tokens.Token(tokens.TYPE_INTEGER, '42')

def test_float():
    assert py2toml.create_primitive_token(4.2) == tokens.Token(tokens.TYPE_FLOAT, '4.2')

def test_bool():
    assert py2toml.create_primitive_token(False) == tokens.Token(tokens.TYPE_BOOLEAN, 'false')
    assert py2toml.create_primitive_token(True) == tokens.Token(tokens.TYPE_BOOLEAN, 'true')

def test_date():
    ts = strict_rfc3339.rfc3339_to_timestamp('1979-05-27T00:32:00-07:00')
    dt = datetime.datetime.fromtimestamp(ts)
    assert py2toml.create_primitive_token(dt) == tokens.Token(tokens.TYPE_DATE, '1979-05-27T07:32:00Z')



from contoml.file.entries import TableEntry, EntryName, AnonymousTableEntry
from contoml.file.cascadedict import CascadeDict


class NamedDict(dict):
    """
    A dict that can use EntryName instances as keys.
    """

    def __init__(self, other_dict=None):
        dict.__init__(self)
        if other_dict:
            for k, v in other_dict.items():
                self[k] = v

    def __setitem__(self, key, value):
        """
        key can be an EntryName instance.

        When key is a path in the form of an EntryName instance, all the parents and grandparents of the value are
        created along the way as instances of NamedDict. If the parent of the value exists, it is replaced with a
        CascadeDict() that cascades the old parent value with a new NamedDict that contains the given child name
        and value.
        """
        if isinstance(key, EntryName):

            if len(key.sub_names) == 1:
                name = key.sub_names[0]
                if name in self:
                    self[name] = CascadeDict(self[name], value)
                else:
                    self[name] = value
            elif len(key.sub_names) == 2:
                parent, child = key.sub_names
                if parent in self:
                    self[parent] = CascadeDict(self[parent], NamedDict({child: value}))
                else:
                    self[parent] = NamedDict({child: value})
            else:
                grand_parents_name, parent_name, child_name = key.sub_names[:-2], key.sub_names[-2], key.sub_names[-1]

                # Make sure parents exist, or create as instances of NamedDict
                d = self
                for name in grand_parents_name:
                    if name not in d:
                        d[name] = NamedDict()
                    d = d[name]

                if parent_name in d:
                    d[parent_name] = CascadeDict(d[parent_name], NamedDict({child_name: value}))
                else:
                    d[parent_name] = NamedDict({child_name: value})
        else:
            return dict.__setitem__(self, key, value)

    def __contains__(self, item):
        try:
            _ = self[item]
            return True
        except KeyError:
            return False

    def append(self, key, value):
        """
        Makes sure the value pointed to by key exists and is a list and appends the given value to it.
        """
        if key in self:
            self[key].append(value)
        else:
            self[key] = [value]

    def __getitem__(self, item):

        if isinstance(item, EntryName):
            d = self
            for name in item.sub_names:
                d = d[name]
            return d
        else:
            return dict.__getitem__(self, item)


def structure(entries):
    """
    Accepts an ordered sequence of Entry instances and returns a navigable object structure representation of the
    TOML file.
    """

    obj = NamedDict()

    last_array_of_tables = None         # The EntryName of the last array-of-tables header

    for entry in entries:

        if isinstance(entry, AnonymousTableEntry):
            obj[''] = entry.table_element

        elif isinstance(entry, TableEntry):
            if last_array_of_tables and entry.name.is_prefixed_with(last_array_of_tables):
                seq = obj[last_array_of_tables]
                unprefixed_name = entry.name.without_prefix(last_array_of_tables)

                seq[-1] = CascadeDict(seq[-1], NamedDict({unprefixed_name: entry.table_element}))
            else:
                obj[entry.name] = entry.table_element
        else:    # It's an ArrayOfTablesEntry

            if last_array_of_tables and entry.name != last_array_of_tables and \
                    entry.name.is_prefixed_with(last_array_of_tables):

                seq = obj[last_array_of_tables]
                unprefixed_name = entry.name.without_prefix(last_array_of_tables)

                if unprefixed_name in seq[-1]:
                    seq[-1][unprefixed_name].append(entry.table_element)
                else:
                    cascaded_with = NamedDict({unprefixed_name: [entry.table_element]})
                    seq[-1] = CascadeDict(seq[-1], cascaded_with)

            else:
                obj.append(entry.name, entry.table_element)
                last_array_of_tables = entry.name

    return obj



from contoml import lexer, parser
from contoml.file import entries
from contoml.parser import elementsanitizer
from contoml.file.entries import EntryName
from contoml.file.structurer import NamedDict, structure
from contoml.parser.tokenstream import TokenStream
from contoml.file.cascadedict import CascadeDict


def test_NamedDict():

    d = NamedDict()

    d[EntryName(('super', 'sub1', 'sub2', 'sub3'))] = 12
    d[EntryName(('super', 'sub1', 'sub2', 'sub4'))] = 42

    assert d[EntryName(('super', 'sub1', 'sub2', 'sub3'))] == 12
    assert isinstance(d[EntryName(('super', 'sub1', 'sub2'))], CascadeDict)
    assert d[EntryName(('super', 'sub1', 'sub2', 'sub4'))] == 42


def test_structure():
    tokens = TokenStream(lexer.tokenize(open('sample.toml').read()))
    elements = elementsanitizer.sanitize(parser.parse_token_stream(tokens))
    entries_ = tuple(entries.extract(elements))

    s = structure(entries_)

    assert s['']['title'] == 'TOML Example'
    assert s['owner']['name'] == 'Tom Preston-Werner'
    assert s['database']['ports'][1] == 8001
    assert s['servers']['alpha']['dc'] == 'eqdc10'
    assert s['clients']['data'][1][0] == 1
    assert s['clients']['key3'] == 'The quick brown fox jumps over the lazy dog.'

    assert s['fruit'][0]['name'] == 'apple'
    assert s['fruit'][0]['physical']['color'] == 'red'
    assert s['fruit'][0]['physical']['shape'] == 'round'
    assert s['fruit'][0]['variety'][0]['name'] == 'red delicious'
    assert s['fruit'][0]['variety'][1]['name'] == 'granny smith'

    assert s['fruit'][1]['name'] == 'banana'
    assert s['fruit'][1]['variety'][0]['name'] == 'plantain'
    assert s['fruit'][1]['variety'][0]['points'][2]['y'] == 4



from contoml.elements.abstracttable import AbstractTable


def to_raw(x):
    from contoml.file.cascadedict import CascadeDict

    if isinstance(x, AbstractTable):
        return x.primitive_value
    elif isinstance(x, CascadeDict):
        return x.neutralized
    elif isinstance(x, (list, tuple)):
        return [to_raw(y) for y in x]
    elif isinstance(x, dict):
        return {k: to_raw(v) for (k, v) in x.items()}
    else:
        return x



from collections import OrderedDict
from contoml.elements import factory
from contoml.elements.array import ArrayElement
from contoml.elements.atomic import AtomicElement
from contoml.elements.inlinetable import InlineTableElement


def test_creating_elements():

    atomic = factory.create_element(42)
    assert isinstance(atomic, AtomicElement)
    assert atomic.value == 42

    seq = factory.create_element(['a', 'p', 'p', 'l', 'e'])
    assert isinstance(seq, ArrayElement)
    assert seq.serialized() == '[a, p, p, l, e]'
    assert ''.join(seq.primitive_value) == 'apple'

    mapping = factory.create_element(OrderedDict((('one', 1), ('two', 2))))
    assert isinstance(mapping, InlineTableElement)
    assert mapping.serialized() == '{one = 1, two = 2}'






class TOMLError(Exception):
    """
    All errors raised by this module are descendants of this type.
    """

class InvalidTOMLFileError(TOMLError):
    pass

class NoArrayFoundError(TOMLError):
    """
    An array of tables was requested but none exist by the given name.
    """

class InvalidValueError(TOMLError):
    pass



from contoml import lexer
from contoml.elements.metadata import WhitespaceElement, NewlineElement, CommentElement, PunctuationElement


def test_whitespace_element():
    element = WhitespaceElement(tuple(lexer.tokenize(' \t   ')))
    assert element.serialized() == ' \t   '


def test_newline_element():
    element = NewlineElement(tuple(lexer.tokenize('\n\n\n')))
    assert element.serialized() == '\n\n\n'


def test_comment_element():
    element = CommentElement(tuple(lexer.tokenize('# This is my insightful remark\n')))
    assert element.serialized() == '# This is my insightful remark\n'


def test_punctuation_element():
    PunctuationElement(tuple(lexer.tokenize('[')))
    PunctuationElement(tuple(lexer.tokenize('[[')))
    PunctuationElement(tuple(lexer.tokenize('.')))
    PunctuationElement(tuple(lexer.tokenize(']')))
    PunctuationElement(tuple(lexer.tokenize(']]')))



from contoml import lexer
from contoml.elements.atomic import AtomicElement
from contoml.elements.metadata import WhitespaceElement, PunctuationElement, NewlineElement, CommentElement
from contoml.elements.table import TableElement


def test_table():

    initial_toml = """name = first
id=42 # My id


"""

    tokens = tuple(lexer.tokenize(initial_toml))

    elements = (
        AtomicElement(tokens[:1]),
        WhitespaceElement(tokens[1:2]),
        PunctuationElement(tokens[2:3]),
        WhitespaceElement(tokens[3:4]),
        AtomicElement(tokens[4:5]),
        NewlineElement(tokens[5:6]),

        AtomicElement(tokens[6:7]),
        PunctuationElement(tokens[7:8]),
        AtomicElement(tokens[8:9]),
        WhitespaceElement(tokens[9:10]),
        CommentElement(tokens[10:12]),

        NewlineElement(tokens[12:13]),
        NewlineElement(tokens[13:14]),
    )

    table = TableElement(elements)

    assert set(table.items()) == {('name', 'first'), ('id', 42)}

    assert table['name'] == 'first'
    assert table['id'] == 42

    table['relation'] = 'another'

    assert set(table.items()) == {('name', 'first'), ('id', 42), ('relation', 'another')}

    table['name'] = 'fawzy'

    assert set(table.items()) == {('name', 'fawzy'), ('id', 42), ('relation', 'another')}

    expected_toml = """name = fawzy
id=42 # My id
relation = another


"""

    assert table.serialized() == expected_toml





from contoml import tokens
from contoml.elements import common
from contoml.elements.metadata import PunctuationElement, NewlineElement, CommentElement


class TraversalMixin:
    """
    A mix-in that provides convenient sub-element traversal to any class with
    an `elements` member that is a sequence of Element instances
    """

    def __find_following_element(self, index, predicate):
        """
        Finds and returns the index of element in self.elements that evaluates the given predicate to True
        and whose index is higher than the given index, or returns -Infinity on failure.
        """
        for i, element in tuple(enumerate(self.elements))[index+1:]:
            if predicate(element):
                return i
        return float('-inf')

    def __find_preceding_element(self, index, predicate):
        """
        Finds and returns the index of the element in self.elements that evaluates the given predicate to True
        and whose index is lower than the given index.
        """

        for i, element in reversed(tuple(enumerate(self.elements))[:index]):
            if predicate(element):
                return i
        return float('-inf')

    def __must_find_following_element(self, predicate):
        """
        Finds and returns the index to the element in self.elements that evaluatest the predicate to True, or raises
        an error.
        """
        i = self.__find_following_element(-1, predicate)
        if i < 0:
            raise RuntimeError('Could not find non-optional element')
        return i

    def _enumerate_non_metadata_sub_elements(self):
        """
        Returns a sequence of of (index, sub_element) of the non-metadata sub-elements.
        """
        return ((i, element) for i, element in enumerate(self.elements) if element.type != common.TYPE_METADATA)

    def _find_preceding_comma(self, index):
        """
        Returns the index of the preceding comma element to the given index, or -Infinity.
        """
        def predicate(element):
            return isinstance(element, PunctuationElement) and element.token.type == tokens.TYPE_OP_COMMA
        return self.__find_preceding_element(index, predicate)

    def _find_following_comma(self, index):
        """
        Returns the index of the following comma element after the given index, or -Infinity.
        """
        def predicate(element):
            return isinstance(element, PunctuationElement) and element.token.type == tokens.TYPE_OP_COMMA
        return self.__find_following_element(index, predicate)

    def _find_following_newline(self, index):
        """
        Returns the index of the following newline element after the given index, or -Infinity.
        """
        return self.__find_following_element(index, lambda e: isinstance(e, NewlineElement))

    def _find_following_comment(self, index):
        """
        Returns the index of the following comment element after the given index, or -Infinity.
        """
        return self.__find_following_element(index, lambda e: isinstance(e, CommentElement))

    def _find_following_line_terminator(self, index):
        """
        Returns the index of the following comment or newline element after the given index, or -Infinity.
        """
        following_comment = self._find_following_comment(index)
        following_newline = self._find_following_newline(index)

        if following_comment == float('-inf'):
            return following_newline
        if following_newline == float('inf'):
            return following_comment

        if following_newline < following_comment:
            return following_newline
        else:
            return following_comment

    def _find_preceding_newline(self, index):
        """
        Returns the index of the preceding newline element to the given index, or -Infinity.
        """
        return self.__find_preceding_element(index, lambda e: isinstance(e, NewlineElement))

    def _find_following_non_metadata(self, index):
        """
        Returns the index to the following non-metadata element after the given index, or -Infinity.
        """
        return self.__find_following_element(index, lambda e: e.type != common.TYPE_METADATA)

    def _find_closing_square_bracket(self):
        """
        Returns the index to the closing square bracket, or raises an Error.
        """
        def predicate(element):
            return isinstance(element, PunctuationElement) and element.token.type == tokens.TYPE_OP_SQUARE_RIGHT_BRACKET
        return self.__must_find_following_element(predicate)

    def _find_following_opening_square_bracket(self, index):
        """
        Returns the index to the opening square bracket, or -Infinity.
        """
        def predicate(element):
            return isinstance(element, PunctuationElement) and element.token.type == tokens.TYPE_OP_SQUARE_LEFT_BRACKET
        return self.__find_following_element(index, predicate)

    def _find_following_table(self, index):
        """
        Returns the index to the next TableElement after the specified index, or -Infinity.
        """
        from contoml.elements.table import TableElement
        return self.__find_following_element(index, lambda e: isinstance(e, TableElement))

    def _find_preceding_table(self, index):
        """
        Returns the index to the preceding TableElement to the specified index, or -Infinity.
        """
        from contoml.elements.table import TableElement
        return self.__find_preceding_element(index, lambda e: isinstance(e, TableElement))

    def _find_closing_curly_bracket(self):
        """
        Returns the index to the closing curly bracket, or raises an Error.
        """
        def predicate(element):
            return isinstance(element, PunctuationElement) and element.token.type == tokens.TYPE_OP_CURLY_RIGHT_BRACKET
        return self.__must_find_following_element(predicate)

    def _find_following_table_header(self, index):
        """
        Returns the index to the table header after the given element index, or -Infinity.
        """
        from contoml.elements.tableheader import TableHeaderElement
        return self.__find_following_element(index, lambda e: isinstance(e, TableHeaderElement))



from setuptools import setup, find_packages

setup(
    name='contoml',
    packages=find_packages(),
    version='0.15',
    description='Consistent TOML for Python',
    author='Amr Hassan',
    author_email='amr.hassan@gmail.com',
    url='https://github.com/Jumpscale/python-consistent-toml',
    keywords=['toml'],
    classifiers=[
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4'
    ],
    install_requires=[
        'six',
        'strict_rfc3339',
        'pytz',
        'timestamp'
    ]
)



from contoml.elements.common import ContainerElement
from contoml.elements import traversal


class AbstractTable(ContainerElement, traversal.TraversalMixin):
    """
    Common code for handling tables as key-value pairs with metadata elements sprinkled all over.

    Assumes input sub_elements are correct.
    """

    def __init__(self, sub_elements):
        ContainerElement.__init__(self, sub_elements)
        self._fallback = None

    def _enumerate_items(self):
        """
        Returns ((key_index, key_element), (value_index, value_element)) for all the element key-value pairs.
        """
        non_metadata = self._enumerate_non_metadata_sub_elements()
        while True:
            yield next(non_metadata), next(non_metadata)

    def items(self):
        for (key_i, key), (value_i, value) in self._enumerate_items():
            yield key.value, value.value

    def keys(self):
        for key, _ in self.items():
            yield key

    def __len__(self):
        return len(tuple(self._enumerate_items()))

    def __contains__(self, item):
        return item in self.keys()

    def _find_key_and_value(self, key):
        """
        Returns (key_i, value_i) corresponding to the given key value.

        Raises KeyError if no matching key found.
        """
        for (key_i, key_element), (value_i, value_element) in self._enumerate_items():
            if key_element.value == key:
                return key_i, value_i
        raise KeyError

    def __getitem__(self, item):
        for key, value in self.items():
            if key == item:
                return value
        raise KeyError

    @property
    def primitive_value(self):
        """
        Returns a primitive Python value without any formatting or markup metadata.
        """
        return {
            key:
                value.primitive_value if hasattr(value, 'primitive_value') else value for key, value in self.items()
            }
